from datetime import datetime, timedelta
from enum import Enum
from peewee import (
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKeyField,
    Model,
    SqliteDatabase,
    TextField,
)

db = SqliteDatabase(
    "tasks.db",
    pragmas={
        "journal_mode": "wal",
        "synchronous": "normal",
        "foreign_keys": 1,
    },
)


class TaskStatus(Enum):
    # order is used for progression in toggle
    ZERO = "zero"
    WIP = "wip"
    BLOCKED = "blocked"
    DONE = "done"


class BaseModel(Model):
    class Meta:
        database = db


class Category(BaseModel):
    name = CharField(unique=True)

    def __str__(self):
        return self.name


class Task(BaseModel):
    # TODO: rename name
    text = TextField()
    status = CharField(
        choices=[(status.value, status.name) for status in TaskStatus],
        default=TaskStatus.ZERO.value,
    )
    due = DateTimeField(null=True)
    category = ForeignKeyField(Category, backref="tasks", null=True)
    type = CharField()
    created_at = DateTimeField(default=datetime.now)
    updated_at = DateTimeField(default=datetime.now)
    deleted = BooleanField(default=False)

    def save(self, *args, **kwargs):
        self.updated_at = datetime.now()
        return super(Task, self).save(*args, **kwargs)


class SavedSearch(BaseModel):
    name = CharField(unique=True)
    filters = CharField()
    sort_string = CharField()

    def __str__(self):
        return self.name


class TaskGenerator(Model):
    template = CharField()
    type = CharField()
    config = TextField()
    deleted = BooleanField(default=False)
    last_generated_at = DateTimeField(null=True)
    created_at = DateTimeField(default=datetime.now)

    def should_generate(self) -> bool:
        """
        generator types: config keys
            recurring: days_between
            monthly: day_of_month
        """
        if self.deleted:
            return False
        if not self.last_generated_at:
            return True

        now = datetime.now()
        if self.type == "recurring":
            days_between = self.config["days_between"]
            days_since = (now - self.last_generated_at).days
            return days_since >= days_between

        elif self.type == "monthly":
            day_of_month = self.generator_config["day_of_month"]

            # check each day until now to see if target day occurred
            one_day = timedelta(days=1)
            check_date = self.last_generated_at + one_day
            while check_date <= now:
                if check_date.day == day_of_month:
                    return True
                check_date += one_day

            return False

        return False


def initialize_db():
    db.connect()
    db.create_tables(
        [
            Category,
            Task,
            SavedSearch,
            # TaskGenerator
        ]
    )
    if not Category.select().exists():
        Category.create(name="default")
    db.close()


if __name__ == "__main__":
    initialize_db()
