import json
from textual.widgets import Input
from datetime import datetime

from ..controller import (
    get_task,
    get_tasks,
    add_task,
    update_task,
    TaskStatus,
    save_view,
    get_saved_view,
)
from ..utils import (
    get_colored_category,
    get_colored_status,
    get_colored_date,
)
from .editor import (
    TableEditor,
    TableColumnConfig,
    NotifyValidationError,
)


def due_preprocessor(val):
    try:
        return datetime.strptime(val, "%Y-%m-%d")
    except ValueError:
        raise NotifyValidationError("Invalid date format. Use YYYY-MM-DD")


def status_preprocessor(val):
    try:
        TaskStatus(val)
        return val
    except ValueError:
        raise NotifyValidationError(
            f"Invalid status. Use: {[s.value for s in TaskStatus]}"
        )


class TT(TableEditor):
    TABLE_CONFIG = (
        TableColumnConfig("id", "ID"),
        TableColumnConfig("text", "Task", default="new task", enable_editor=True),
        TableColumnConfig(
            "status",
            "Status",
            default="zero",
            enum=TaskStatus,
            preprocessor=status_preprocessor,
        ),
        TableColumnConfig("type", "Type", default=""),
        TableColumnConfig("due", "Due", default="", preprocessor=due_preprocessor),
        TableColumnConfig("category", "Category", default="main"),
    )

    BINDINGS = [
        # saved views
        ("ctrl+s", "save_view", "save current view"),
        ("ctrl+o", "load_view", "load saved view"),
    ]

    def __init__(self, default_view="default"):
        super().__init__()
        self._load_view(default_view)
        self.update_item_callback = update_task
        self.add_item_callback = add_task
        self.get_item_callback = get_task

    def _load_view(self, name):
        try:
            saved = get_saved_view(name)
            self.filters = json.loads(saved.filters)
            self.sort_string = saved.sort_string
        except Exception:
            self.notify(f"Could not load {name}")

    def action_save_view(self):
        self._show_input("save-view", "default")

    def action_load_view(self):
        self._show_input("load-view", "")

    def on_input_submitted(self, event: Input.Submitted):
        # Override to add save/load view
        if self.mode == "save-view":
            save_view(event.value, filters=self.filters, sort_string=self.sort_string)
        elif self.mode == "load-view":
            self._load_view(event.value)
            self.refresh_tasks(restore_cursor=False)
        else:
            super().on_input_submitted(event)
            return

    def refresh_items(self):
        items = get_tasks(
            self.search_query,
            category=self.filters.get("category"),
            statuses=self.filters.get("status", "").split(",")
            if "status" in self.filters
            else None,
            sort=self.sort_string,
        )
        for item in items:
            category = get_colored_category(
                item.category.name if item.category else " - "
            )
            status = get_colored_status(item.status)
            due = get_colored_date(item.due)

            self.table.add_row(
                str(item.id),
                item.text.split("\n")[0],  # first line
                status,
                item.type,
                due,
                category,
                key=str(item.id),
            )


def run(default_view):
    app = TT(default_view)
    app.run()


if __name__ == "__main__":
    run()
